"""

Core of onnxexplorer

"""
from typing import Callable, List, Dict, Union, Optional, Tuple, Sequence, TypeVar
import onnx
import onnxruntime as ort
from onnx import AttributeProto, TensorProto, GraphProto, ModelProto, NodeProto
from typing import Dict, Optional, Union, cast, Text, IO
from google import protobuf
from colorama import init, Fore, Style, Back
import os
import sys
import numpy as np
from tabulate import tabulate
from rich.console import Console
from rich.columns import Columns
from rich.panel import Panel
from rich.table import Table
from rich import box
from .common import elem_type_to_name

def load_onnx_model(f, proto):
    s = None
    if hasattr(f, 'read') and callable(cast(IO[bytes], f).read):
        s = cast(IO[bytes], f).read()
    else:
        with open(cast(Text, f), 'rb') as readable:
            s = readable.read()
    decoded = cast(Optional[int], proto.ParseFromString(s))
    if decoded is not None and decoded != len(s):
        raise protobuf.message.DecodeError(
            "Protobuf decoding consumed too few bytes: {} out of {}".format(
                decoded, len(s)))
    return proto


def search_nodes_by_type(model, type_name):
    """
    search all nodes that with Type
    :param model_proto:
    :param type_name:
    :return:
    """
    print(Style.BRIGHT + 'search node by op type: ' +
          Style.RESET_ALL + Fore.GREEN + type_name + Style.RESET_ALL)
    if isinstance(model, ModelProto):
        graph = model.graph
        i = 0
        for node in graph.node:
            if type_name in node.op_type:
                i += 1
                print(node)
        print(Style.BRIGHT + 'listed all {} {} nodes in detail.\n'.format(i,
                                                                          type_name) + Style.RESET_ALL)


def list_nodes(model):
    """
    print out all nodes
    :param model_proto:
    :return:
    """
    print(Style.BRIGHT + 'start list nodes names:' + Style.RESET_ALL)
    if isinstance(model, ModelProto):
        graph = model.graph
        i = 0
        for node in graph.node:
            i += 1
            print('{}. {}'.format(i, node.name))
        print(Style.BRIGHT + 'listed all {} nodes names.\n'.format(i) + Style.RESET_ALL)


def list_nodes_hl(model):
    print(Style.BRIGHT + 'start list nodes all:' + Style.RESET_ALL)
    if isinstance(model, ModelProto):
        graph = model.graph
        i = 0
        for node in graph.node:
            if len(node.attribute) > 0 and node.attribute[0].t.raw_data != None:
                data_type = node.attribute[0].t.data_type
                if data_type == 7:
                    raw_data = np.frombuffer(
                        node.attribute[0].t.raw_data, dtype=np.int64)
                elif data_type == 1:
                    raw_data = np.frombuffer(
                        node.attribute[0].t.raw_data, dtype=np.float32)
                else:
                    raw_data = np.frombuffer(
                        node.attribute[0].t.raw_data, dtype=np.int64)
                print('|raw data that encoded inside this node: {}|'.format(raw_data))
            print(node)
            i += 1
        print(Style.BRIGHT +
              'listed all {} nodes in detail.\n'.format(i) + Style.RESET_ALL)


def search_node_by_id(model, node_id):
    """
    print out specific node by ID
    :param model_proto:
    :param node_id:
    :return:
    """
    print(Style.BRIGHT + 'search node by ID: ' +
          Style.RESET_ALL + Fore.GREEN + node_id + Style.RESET_ALL)
    if isinstance(model, ModelProto):
        graph = model.graph
        i = 0
        for node in graph.node:
            if node_id in node.name:
                i += 1
                print(node)
        print(Style.BRIGHT +
              'listed all {} nodes in detail.\n'.format(i) + Style.RESET_ALL)


def get_all_used_op_types(model):
    if isinstance(model, ModelProto):
        graph = model.graph
        all_op_types = {}
        for node in graph.node:
            if node.op_type not in all_op_types.keys():
                all_op_types[node.op_type] = 1
            else:
                all_op_types[node.op_type] += 1
    return all_op_types


def get_input_and_output_shapes(model, verbose=False) -> Union[Dict[str, list], None]:
    input_specs = dict()
    output_specs = dict()

    _inputs = get_inputs(model)
    for _input in _inputs:
        tensor_type = _input.type.tensor_type
        sh = []
        # check if it has a shape:
        if (tensor_type.HasField("shape")):
            # iterate through dimensions of the shape:
            for d in tensor_type.shape.dim:
                # the dimension may have a definite (integer) value or a symbolic identifier or neither:
                if (d.HasField("dim_value")):
                    sh.append(d.dim_value)
                    if verbose:
                        print(d.dim_value, end=", ")  # known dimension
                elif (d.HasField("dim_param")):
                    sh.append(-1)
                    # unknown dimension with symbolic name
                    if verbose:
                        print(d.dim_param, end=", ")
                else:
                    sh.append('?')
                    if verbose:
                        print("?", end=", ")  # unknown dimension with no name
            input_specs[_input.name] = (sh, elem_type_to_name(tensor_type.elem_type))
        else:
            print("unknown rank", end="")

    for output in model.graph.output:
        if verbose:
            print(output.name, end=": ")
        # get type of input tensor
        tensor_type = output.type.tensor_type
        sh = []
        # check if it has a shape:
        if (tensor_type.HasField("shape")):
            # iterate through dimensions of the shape:
            for d in tensor_type.shape.dim:
                # the dimension may have a definite (integer) value or a symbolic identifier or neither:
                if (d.HasField("dim_value")):
                    sh.append(d.dim_value)
                    if verbose:
                        print(d.dim_value, end=", ")  # known dimension
                elif (d.HasField("dim_param")):
                    sh.append(-1)
                    # unknown dimension with symbolic name
                    if verbose:
                        print(d.dim_param, end=", ")
                else:
                    sh.append('?')
                    if verbose:
                        print("?", end=", ")  # unknown dimension with no name
            output_specs[output.name] = (sh, elem_type_to_name(tensor_type.elem_type))
        else:
            print("unknown rank", end="")

    return input_specs, output_specs


def summary(model_proto, model_p, model_name, verbose=False):
    if isinstance(model_proto, ModelProto):
        all_ops = get_all_used_op_types(model_proto)
        opi_vd = ''
        for opi in model_proto.opset_import:
            opi_vd += str(opi.version)
            opi_vd += opi.domain
            opi_vd += ', '
        summary_str = f'IR Version: {model_proto.ir_version}\n' + \
            f'Opset Version: {opi_vd}\n' + \
            f'Doc: {model_proto.doc_string}\n' + \
            f'Producer Name: {model_proto.producer_name}\n' + \
            f"All Ops: {','.join(all_ops.keys())}"

        console = Console()
        p = Panel(summary_str,
                  title=f'{model_name} Summary', highlight=True)
        console.print(p, width=90)

        inp_specs, oup_specs = get_input_and_output_shapes(
            model_proto, verbose=verbose)
        inp_specs = {k: (v[0], 'input', v[1]) for k, v in inp_specs.items()}
        oup_specs = {k: (v[0], 'output', v[1]) for k, v in oup_specs.items()}
        inp_specs.update(oup_specs)

        # print table

        table = Table(title=f'{model_name} Detail',
            caption='Table generated by onnxexplorer', expand=True,
            row_styles=["none", "none"],
            pad_edge=True,
            box=box.ROUNDED,)
        table.add_column('[yellow]Name')
        table.add_column('[cyan]Shape')
        table.add_column('[blue]Input/Output')
        table.add_column('[green]Dtype')
        for k, v in inp_specs.items():
            table.add_row(k, str(v[0]), v[1], v[2])
        # panel_table = Panel(table, title=f'{model_name} Detail',)
        # console.print(panel_table)
        console.print(table, width=90)

        if verbose:
            # using table to show
            stat_data = []
            a_nodes = 0
            for k, v in all_ops.items():
                stat_data.append([k, v])
                a_nodes += v
            print(Fore.BLUE + Style.BRIGHT +
                  '\nStatistic table: ' + Style.RESET_ALL)
            print(tabulate(stat_data, headers=[
                'Op Name', 'Number']))
            print('Total {} kinds ops, {} nodes.'.format(len(stat_data), a_nodes))


def get_inputs(model: onnx.ModelProto) -> List[onnx.ValueInfoProto]:
    initializer_names = [x.name for x in model.graph.initializer]
    return [ipt for ipt in model.graph.input if ipt.name not in initializer_names]


def get_input_names(model: onnx.ModelProto) -> List[str]:
    input_names = [ipt.name for ipt in get_inputs(model)]
    return input_names



from onnx import mapping


def elem_type_to_name(elem_type: int):
    type_map = {
        1: 'float32',
        2: 'uint8',
        3: 'int8',
        4: 'uint16',
        5: 'int16',
        6: 'int32',
        7: 'int64',
        8: 'string',
        9: 'boolean',
        10: 'float16',
        11: 'float64',
        12: 'uint32',
        13: 'uint64',
        14: '',
        15: 'float32',
        16: 'float32',
    }
    type_map = mapping.TENSOR_TYPE_TO_NP_TYPE
    if elem_type in type_map.keys():
        return type_map[elem_type].name
    else:
        return 'undefined'


"""

This script provides
various exploration on onnx model
such as those operations:

- summary:
    this command will summarize model info mation such as:
    1. model opset version;
    2. if check pass;
    3. whether can be simplifiered or not;
    4. inputs node and outputs node;
    5. all nodes number;
    6. Initializers tensornumber and their index;

- search:
    1. search all nodes by OpType;
    2. search node by node name (ID);

- list:
    1. print out all nodes id
    2. list -hl will print all nodes and it's attribute

"""
from datetime import datetime
from .version import __version__, short_version
import os
import sys
import argparse
from tabnanny import verbose
import onnx
from onnx import ModelProto, TensorProto, NodeProto
from typing import Optional, cast, Text, IO
from google import protobuf
from colorama import init, Fore, Style, Back
from .core import *
from alfred.utils.log import logger as logging
from .msg import print_welcome_msg

try:
    from .totrt import convert_onnx_to_tensorrt
except ImportError:
    logging.warning('disable TensorRT since it was not found.')


try:
    import onnxruntime as ort
except ImportError:
    logging.error(
        "onnxexp needs onnxruntime, you need install onnxruntime first.")
    exit(0)

init()


dt = datetime.now()


def arg_parse():
    """
    parse arguments
    :return:
    """
    parser = argparse.ArgumentParser(prog="onnxexp")
    parser.add_argument(
        "--version", "-v", action="store_true", help="show version info."
    )

    main_sub_parser = parser.add_subparsers(dest="subparser_name")

    # =============== glance part ================
    vision_parser = main_sub_parser.add_parser(
        "glance", help="Take a glance at your onnx model."
    )
    vision_parser.add_argument("--model", "-m", help="onnx model path")
    vision_parser.add_argument(
        "--verbose", "-v", action="store_true", help="verbose info"
    )

    # =============== totrt part ================
    trt_parser = main_sub_parser.add_parser(
        "totrt", help="Convert your onnx model to tensorrt engine."
    )
    trt_parser.add_argument("--model", "-m", help="onnx model path")
    trt_parser.add_argument(
        "--data_type", "-d", type=int, default=32, help="32, 16, 8 presents fp32, fp16, int8"
    )
    trt_parser.add_argument(
        "--batch_size", "-b", type=int, default=1, help="batch size of your model."
    )
    trt_parser.add_argument("--min_shapes", nargs='+',
                            help="min_shapes of opt_params")
    trt_parser.add_argument("--opt_shapes", nargs='+',
                            help="opt_shapes of opt_params")
    trt_parser.add_argument("--max_shapes", nargs='+',
                            help="max_shapes of opt_params")
    trt_parser.add_argument('--suffix', action='store_true',
                            help='if enable, generated engine will have a suffix.')

    # =============== check part ================
    check_parser = main_sub_parser.add_parser(
        "check", help="Check your onnx model is valid or not."
    )
    check_parser.add_argument("--model", "-m", help="onnx model path")
    check_parser.add_argument("--print", action="store_true")

    return parser.parse_args()


class ONNXExplorer(object):
    def __init__(self):
        args = arg_parse()
        if args.version:
            print_welcome_msg(version=__version__)
        else:
            if args.subparser_name == None:
                print("should provide at least one sub command, -h for detail.")
                exit(-1)
            self.model_path = args.model
            if args.model != None and os.path.exists(args.model):
                print(
                    Style.BRIGHT
                    + "Exploring on onnx model: "
                    + Style.RESET_ALL
                    + Fore.GREEN
                    + self.model_path
                    + Style.RESET_ALL
                )
                # self.model_proto = load_onnx_model(self.model_path, ModelProto())
                self.model_proto = onnx.load(self.model_path)
                if args.subparser_name == "glance":
                    summary(self.model_proto, self.model_path,
                            os.path.basename(args.model), args.verbose)
                elif args.subparser_name == "totrt":
                    convert_onnx_to_tensorrt(
                        self.model_path, args.data_type, args.batch_size, args.min_shapes, args.opt_shapes, args.max_shapes, with_suffix=args.suffix)
                elif args.subparser_name == "check":
                    self.check(args.print)

            else:
                print(
                    "{} does not exist or you should provide model path like `onnxexp glance -m model.onnx`.".format(
                        args.model
                    )
                )

    def ls(self):
        parser = argparse.ArgumentParser(
            description="ls at any level about the model.")
        # prefixing the argument with -- means it's optional
        parser.add_argument("-hl", action="store_true")
        args = parser.parse_args(sys.argv[3:])
        print("Running onnxexp ls, hl=%s" % args.hl)
        if args.hl:
            list_nodes_hl(self.model_proto)
        else:
            list_nodes(self.model_proto)

    def search(self):
        parser = argparse.ArgumentParser(
            description="search model node by name or type"
        )
        # NOT prefixing the argument with -- means it's not optional
        parser.add_argument("--name", "-n", help="name of to search node")
        parser.add_argument("--type", "-t", help="type of to search node")
        args = parser.parse_args(sys.argv[3:])
        if args.name != None:
            search_node_by_id(self.model_proto, args.name)
        elif args.type != None:
            search_nodes_by_type(self.model_proto, args.type)
        else:
            print("search should provide type name or id to search.")

    def check(self, p):
        logging.info("checking on model: {}".format(self.model_path))
        logging.info("this command will check an onnx model is valid or not.")
        onnx.checker.check_model(self.model_proto)
        if p:
            a = onnx.helper.printable_graph(self.model_proto.graph)
            print(a)


def main():
    ONNXExplorer()


if __name__ == "__main__":
    main()



from alfred.deploy.tensorrt.common import (
    build_engine_onnx_v2,
    check_engine,
)
from alfred.utils.log import logger
import os


def convert_onnx_to_tensorrt(onnx_f, datatype, batch_size=1, min_shapes=None, opt_shapes=None, max_shapes=None, with_suffix=True):
    opt_params = None
    if min_shapes is not None and opt_shapes is not None and max_shapes is not None:
        names = [i.split(':')[0] for i in min_shapes]
        mins_shapes = [[int(ii) for ii in i.split(':')[1].split('x')]
                       for i in min_shapes]
        opt_shapes = [[int(ii) for ii in i.split(':')[1].split('x')]
                      for i in opt_shapes]
        max_shapes = [[int(ii) for ii in i.split(':')[1].split('x')]
                      for i in max_shapes]

        opt_params = dict()
        for i in range(len(min_shapes)):
            opt_params[names[i]] = [
                mins_shapes[i], opt_shapes[i], max_shapes[i]]
        batch_size = '_dynamic_' + str(max_shapes[0][0])

    fp16 = False
    int8 = False
    if datatype == 16:
        logger.info('enable fp16 mode.')
        fp16 = True
    elif datatype == 8:
        print("int8 not supported.")
    if with_suffix:
        if fp16:
            save_p = os.path.join(os.path.dirname(onnx_f), os.path.basename(
                onnx_f).replace(".onnx", "_fp16_bs{}.engine".format(batch_size)))
        elif int8:
            save_p = os.path.join(os.path.dirname(onnx_f), os.path.basename(
                onnx_f).replace(".onnx", "_int8_bs{}.engine".format(batch_size)))
        else:
            save_p = os.path.join(os.path.dirname(onnx_f), os.path.basename(
                onnx_f).replace(".onnx", "_bs{}.engine".format(batch_size)))
    else:
        save_p = os.path.join(os.path.dirname(onnx_f), os.path.basename(
            onnx_f).replace(".onnx", ".engine"))
    logger.info('engine will be saved into: {}'.format(save_p))
    engine = build_engine_onnx_v2(
        onnx_file_path=onnx_f, max_batch_size=batch_size, engine_file_path=save_p, fp16_mode=fp16, save_engine=True, opt_params=opt_params
    )
    # check_engine(engine)
    logger.info('engine saved into: {}'.format(save_p))



"""
This example shows how to display content in columns.
The data is pulled from https://randomuser.me
"""

import json
from urllib.request import urlopen
from itertools import accumulate
from bisect import bisect
from random import randrange
from unicodedata import name as unicode_name


from rich.console import Console
from rich.columns import Columns
from rich.panel import Panel


UNICODE_VERSION = 6

# Sauce: http://www.unicode.org/charts/PDF/U1F300.pdf
EMOJI_RANGES_UNICODE = {
    6: [
        ('\U0001F300', '\U0001F320'),
        ('\U0001F330', '\U0001F335'),
        ('\U0001F337', '\U0001F37C'),
        ('\U0001F380', '\U0001F393'),
        ('\U0001F3A0', '\U0001F3C4'),
        ('\U0001F3C6', '\U0001F3CA'),
        ('\U0001F3E0', '\U0001F3F0'),
        ('\U0001F400', '\U0001F43E'),
        ('\U0001F440', ),
        ('\U0001F442', '\U0001F4F7'),
        ('\U0001F4F9', '\U0001F4FC'),
        ('\U0001F500', '\U0001F53C'),
        ('\U0001F540', '\U0001F543'),
        ('\U0001F550', '\U0001F567'),
        ('\U0001F5FB', '\U0001F5FF')
    ],
    7: [
        ('\U0001F300', '\U0001F32C'),
        ('\U0001F330', '\U0001F37D'),
        ('\U0001F380', '\U0001F3CE'),
        ('\U0001F3D4', '\U0001F3F7'),
        ('\U0001F400', '\U0001F4FE'),
        ('\U0001F500', '\U0001F54A'),
        ('\U0001F550', '\U0001F579'),
        ('\U0001F57B', '\U0001F5A3'),
        ('\U0001F5A5', '\U0001F5FF')
    ],
    8: [
        ('\U0001F300', '\U0001F579'),
        ('\U0001F57B', '\U0001F5A3'),
        ('\U0001F5A5', '\U0001F5FF')
    ]
}

NO_NAME_ERROR = '(No name found for this codepoint)'


def get_random_emoji(unicode_version=6):
    if unicode_version in EMOJI_RANGES_UNICODE:
        emoji_ranges = EMOJI_RANGES_UNICODE[unicode_version]
    else:
        emoji_ranges = EMOJI_RANGES_UNICODE[-1]

    # Weighted distribution
    count = [ord(r[-1]) - ord(r[0]) + 1 for r in emoji_ranges]
    weight_distr = list(accumulate(count))

    # Get one point in the multiple ranges
    point = randrange(weight_distr[-1])

    # Select the correct range
    emoji_range_idx = bisect(weight_distr, point)
    emoji_range = emoji_ranges[emoji_range_idx]

    # Calculate the index in the selected range
    point_in_range = point
    if emoji_range_idx != 0:
        point_in_range = point - weight_distr[emoji_range_idx - 1]

    # Emoji 😄
    emoji = chr(ord(emoji_range[0]) + point_in_range)
    emoji_name = unicode_name(emoji, NO_NAME_ERROR).capitalize()
    emoji_codepoint = "U+{}".format(hex(ord(emoji))[2:].upper())

    return (emoji, emoji_codepoint, emoji_name)





def get_panel_desc(version):
    p = f'{get_random_emoji()[0]} Author: Lucas Jin\n' + \
        f'{get_random_emoji()[0]} Github: https://github.com/jinfagang/onnxexplorer, please star 🌟\n' + \
        f'{get_random_emoji()[0]} Version: {version}\n' + \
        f'{get_random_emoji()[0]} Since: 2021.10 \n' + \
        f'{get_random_emoji()[0]} Promote: onnxexp glance -m a.onnx'
    return p


def print_welcome_msg(version):
    console = Console()
    p = Panel(get_panel_desc(version),
            title=f'{get_random_emoji()[0]} Welcome to onnxexp! {get_random_emoji()[0]}', highlight=True, expand=True)
    console.print(p, width=90)
